/*
  ICFS: Interactively Controlled File System
  Copyright (C) 2024-2025  Fedir Kovalov

  This program can be distributed under the terms of the GNU GPLv2.
  See the file LICENSE.
*/

#include "access_t.h"
#include <stddef.h>
#include <sys/types.h>
#include <time.h>
#define _GNU_SOURCE
#include "cc.h"
#include "perm_permissions_table.h"
#include "real_filename.h"
#include "temp_permissions_table.h"
#include "ui-socket.h"
#include <pthread.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/socket.h>
#include <sys/un.h>
#include <unistd.h>

#define ZENITY_YES 0
#define ZENITY_NO 1
#define ZENITY_PERM 2

int init_ui_socket(const char *perm_permissions_db_filename) {
  FILE *fp = NULL;

  if (init_temp_permissions_table()) {
    fprintf(stderr, "Could not initialize temporary permissions table.\n");
    return 1;
  }

  if (init_perm_permissions_table(perm_permissions_db_filename)) {
    fprintf(stderr, "Could not initialize permanent permissions table.\n");
    return 1;
  }

  // Test if Zenity is installed (get version)
  fp = popen("zenity --version", "r");
  if (fp == NULL) {
    perror("Pipe returned an error");
    return 1;
  }

  pclose(fp);
  return 0;
}

void destroy_ui_socket(void) {
  destroy_temp_permissions_table();
  destroy_perm_permissions_table();
}

/**
 * Asks the user if the process should be allowed to access the file using the
 * GUI
 *
 * @param filename: The file that the process is trying to access
 * @param pi: The process information
 * @return: access status - ALLOW, DENY or ALLOW_TEMP
 * allowed for the runtime of the process
 */
access_t ask_access(const char *filename, struct process_info proc_info) {
  FILE *fp = NULL;
  char *command = NULL;
  int ret = asprintf(&command, "zenity \"%d\" \"%s\" \"%s\" \"%s\"",
                     proc_info.PID, proc_info.name, filename, get_mountpoint());

  if (ret < 0) {
    // If asprintf fails, the contents of command are undefined (see man
    // asprintf). That does not explicitly rule out that command will be a valid
    // pointer. But the risk of freeing a non-allocated pointer is too much to
    // justify preparing for this.
    fprintf(stderr, "Could not create query on rule insertion");
    perror("");
    return 1;
  }

  // Zenity Question Message Popup
  fp = popen(command, "r");
  free(command);

  if (fp == NULL) {
    perror("Pipe returned a error");
    return DENY;
  }

  str(char) zenity_output;
  init(&zenity_output);

  size_t total_read = 0;
  char line[1024]; // Buffer to read individual lines

  // Read the command output line by line
  while (fgets(line, sizeof(line), fp)) {
    size_t line_len = strlen(line);
    push_fmt(&zenity_output, line);
  }

  int zenity_exit_code = WEXITSTATUS(pclose(fp));
  fprintf(stderr, "zenity wrote out %s\n", first(&zenity_output));
  fprintf(stderr, "zenity returned %d\n", zenity_exit_code);

  cleanup(&zenity_output);

  if (zenity_exit_code == (ZENITY_YES | ZENITY_PERM)) {
    return ALLOW;
  }
  if (zenity_exit_code == ZENITY_YES) {
    return ALLOW_TEMP;
  }

  return DENY;
}

/**
 * Check access according to:
 * 1. temp permission table
 * 2. permanent permission table
 * 3. user descision
 *
 * @param filename: The file that the process is trying to access
 * @pram pi: The process information
 * @param opts: options (GRANT_TEMP, GRANT_PERM)
 * @return: 0 if access is denied, 1 if access is allowed
 */
int interactive_access(const char *filename, struct process_info proc_info,
                       int opts) {
  char *real_path = real_filename(filename);

  access_t access = check_temp_access(real_path, proc_info);
  if (access == ALLOW) {
    free(real_path);
    return 1;
  }
  if (access == DENY) {
    free(real_path);
    return 0;
  }

  access = check_perm_access(real_path, proc_info);
  if (access == ALLOW) {
    free(real_path);
    return 1;
  }
  if (access == DENY) {
    free(real_path);
    return 0;
  }

  // if noth GRANT_TEMP and GRANT_PERM are selected, then only permanent
  // permissions are granted

  if (opts & GRANT_PERM) {
    give_perm_access(real_path, proc_info);
    free(real_path);
    return 1;
  }
  if (opts & GRANT_TEMP) {
    set_temp_access(real_path, proc_info, SET_ALLOW);
    free(real_path);
    return 1;
  }

  access_t user_response = ask_access(real_path, proc_info);
  if (user_response == ALLOW) {
    give_perm_access(real_path, proc_info);
    free(real_path);
    return 1;
  }

  if (user_response == ALLOW_TEMP) {
    set_temp_access(real_path, proc_info, SET_ALLOW);
    free(real_path);
    return 1;
  }

  if (user_response == DENY) {
    set_temp_access(real_path, proc_info, SET_DENY);
    free(real_path);
    return 0;
  }

  free(real_path);
  // deny on unknown options.
  return 0;
}
